// *** Zahnrad-Zeichenprogramm ***
// Umgesetzt aus dem gleichnamigen QuickBASIC-Programm
// (c) 1990, 1997 by Andreas Meile

#include <stdio.h>
#include <conio.h>
#include <string.h>
#include <math.h>
#include <grx.h>
#include "eing_rou.h" // Paket mit syntaxprfenden Eingaberoutinen

const int max_bilder = 300;       // Oberlimit fr die Zahl der
                                  // Einzelbilder bei Animation
const float pi = 4.0 * atan(1.0); // Konstante 
const float pi_2 = 2.0 * pi;      // Wert 2  (360)
const float pi_12 = pi / 2.0;     // Wert /2  (90)
const float pi_32 = 1.5 * pi;     // Wert 3/2 (270)
const float umr = pi / 180.0;     // Umrechnungsfaktor  => rad

const float y_x_verhaeltnis = 3.0 / 4.0; // Bildschirmverhltnis: Bei
// praktisch smtlichen heutigen Bildrhren immer 4:3, sodass in den
// Auflsungen 640x480, 800x600 und 1024x768 quadratische Bildpunkte
// resultieren

int treiber_konst = GR_width_height_color_graphics, // Hardware-Konstante fr
    // Grafikaktivierung; falls mit lterem Treiber ohne Explizit-Modusangabe
    // gearbeitet wird, muss vom Benutzer ein anderer Wert eingegeben werden
    x_aufl = 640,     // horizontale Grafikauflsung
    y_aufl,           // vertikale Grafikauflsung
    a_far = 16,       // Anzahl Farben
    a_bilder,         // Anzahl Bilder fr Animation
    e_fl = 1,         // Flag fr erstes Rad beim Limitencheck
    alter_videomodus; // Zwischenspeicher fr alter Modus

    float x_ba,    // Koordinatenverschiebungsbetrag x
          y_ba,    // Koordinatenverschiebungsbetrag y
          fx,      // Koordinatenumrechnungsfaktor x
          fy,      // Koordinatenumrechnungsfaktor y
          k,       // Pixelverhltnis Hhe/Breite
          x_min,   // linke Ausdehnung
          x_max,   // rechte Ausdehnung
          y_min,   // untere Ausdehnung
          y_max;   // obere Ausdehnung

// Klasse fr Zahnrder
class zahnrad {
  public:
    float lx;     // X-Koordinate des Mittelpunktes (Benutzereinheiten)
    float ly;     // Y-Koordinate des Mittelpunktes (Benutzereinheiten)
    float lal;    // Stellungswinkel des 1. Zahns in 
    float m;      // Modul (Benutzereinheiten)
    int z;        // Zhnezahl
    float x;      // Profilverschiebungsfaktor gemss Norm (0=unkorrigiert)
    float hff;    // Fusshhenfaktor (bei genormten Rder normalerweise 1.25)
    float haf;    // Kopfhhenfaktor (bei genormten Rder normalerweise 1)
    float al;     // Eingriffswinkel in  (bei genormten Rder normalerweise 20)

    // Limiten aktualisieren
    void aktualisiere_limiten();

    // Zeichnen
    void zeichne(int far);     // far = Zeichenfarbe

    // Parameter so berechnen, dass
    void berechne_neulage (const zahnrad &zref, // Bezugszahnrad
                                 float w, // Lagewinkel vom Bezugsrad in 
                                 int nz=0, // Zhnezahl des neuen Rades;
                            // 0=automatisch als V-Null-Verzahnung berechnet
                                 float a=0.0); // Abstand; bei 0 als V-Null-
                                               // Verzahnung berechnet
};

// Umgrenzungen aktualisieren
void zahnrad::aktualisiere_limiten() {
  register float h = m * (ldexp(float(z), -1) + haf + x);
  if(lx - h < x_min || e_fl)
    x_min = lx - h;
  if(lx + h > x_max || e_fl)
    x_max = lx + h;
  if(ly - h < y_min || e_fl)
    y_min = ly - h;
  if(ly + h > y_max || e_fl)
    y_max = ly + h;
  e_fl = 0;
}

// Zahnrad zeichnen
void zahnrad::zeichne(int farb) {
  float px = (lx - x_ba) * fx,
        py = (y_ba - ly) * fy,
        my = m * fy,
        r = ldexp(my * float(z), -1),
        rf = r - my * (hff - x),
        rf2 = rf * rf,
        ra = r + my * (haf + x),
        ra2 = ra * ra,
        lalr = lal * umr,
        alr = al * umr,
        rg = r * cos(alr),
        ky = 1.0 / (rg * rg),
        kx = ky / (k * k),
        teilw = pi_2 / float(z),
        psi = ldexp(teilw, -2) - alr + tan(alr) * (1.0 + ldexp(x, 1) /
        float(z));

  int ryao = int(py - ra),
      ryau = int(py + ra),
      ryfo = int(py - rf),
      ryfu = int(py + rf);

  for(int yy=ryau; yy > ryao; yy--) {
    float yh1 = py - float(yy);
    register float h = k * sqrt(ra2 - yh1 * yh1);
    int xh1 = int(floor(px - h)),
        xh2 = int(floor(px + h)),
        stat = (yy > ryfo && yy <= ryfu ? 2 : 1);
    do {
      int a;
      if(stat == 2) {
        a = xh2;
        h = k * sqrt(rf2 - yh1 * yh1);
        xh2 = int(floor(px - h));
      }
      for(int xx=xh2; xx > xh1; xx--) {
        float xh1 = float(xx) - px,
              phi = (xh1 == 0.0 ?
                ( yh1 > 0.0 ?
                  pi_12
                :
                  pi_32
                )
              :
                atan(k * yh1 / xh1) +
                (xh1 < 0.0 ? pi : 0.0)
              );
        phi -= lalr;
        phi -= teilw * floor(phi / teilw);
        if(phi > ldexp(teilw, -1))
          phi = teilw - phi;
        float d = xh1 * xh1 * kx + yh1 * yh1 * ky;
        if(d < 1.0) {
          if(phi < psi)
            GrPlot(xx, yy, farb);
        } else {
          register float bet = sqrt(d - 1.0);
          if(phi < psi - bet + atan(bet))
            GrPlot(xx, yy, farb);
        }
      }
      if(stat == 2) {
        xh1 = int(floor(px + h));
        if(xh1 > xh2)
          GrHLine(xh2 + 1, xh1, yy, farb);
        xh2 = a;
      }
      stat--;
    } while(stat);
  }
}

// Neues Zahnrad anhand der Daten
void zahnrad::berechne_neulage (const zahnrad &zref, float w, int nz,
float a) {
  // Umkopieren
  m = zref.m;
  hff = zref.hff;
  haf = zref.haf;
  al = zref.al;
  if(a == 0.0) {
    a = float(zref.z + nz) * ldexp(zref.m, -1);
    x = -zref.x;
  } else {
    if(nz == 0)
      nz = int(ldexp(a, 1) / zref.m) - zref.z;
    float h3 = zref.al * umr,
          h1 = zref.m * float(zref.z + nz) * ldexp(cos(h3), -1) / a,
          h2 = sqrt(1.0 - h1 * h1) / h1;
    x = float(zref.z + nz) * ldexp((h2 - atan(h2) + h3) / tan(h3) -
    1.0, -1) - zref.x;
  }
  lx = zref.lx + a * cos(w * umr);
  ly = zref.ly + a * sin(w * umr);
  lal = w - 180.0 + 180.0 / float(nz) + (w - zref.lal) * float(zref.z) /
  float(nz);
  z = nz;
}

// Koordinatenbereich setzen
void setze_bereich (float xi,   // linke Limite
                    float xa,   // rechte Limite
                    float yi,   // untere Limite
                    float ya) { // obere Limite

  fx = float(x_aufl) / (xa - xi);
  fy = float(y_aufl) / (ya - yi);
  k = fx / fy;
  x_ba = xi + 0.5 / fx;
  y_ba = ya - 0.5 / fy;
}

const int n_zahnraeder_max = 50;

zahnrad zr[n_zahnraeder_max];   // Speicher fr Zahnrder

int omega[n_zahnraeder_max],    // Geschwindigkeiten
    farbe[n_zahnraeder_max];    // Farben

void main() {
  char d_name[80];
  FILE *d_handle;

  float x_min1,   // Bildschirmbezogene Ausdehnung links
        x_max1,   // dito rechts
        y_min1,   // dito unten
        y_max1;   // dito oben

  printf("\033[0m\033[2J\033[2;1HZahnradzeichenprogramm mit Animation\n"
         "(c) 1990, 1997 by Andreas Meile, CH-8242 Hofen SH\n\n"
         "Datei mit Vorgelegedefinition? ");
  gets(d_name);

  // Standardwerte fr das erste Rad
  int zi = 0;
  zr[zi].lx = 0.0;
  zr[zi].ly = 0.0;
  zr[zi].lal = 0.0;
  zr[zi].m = 1.0;
  zr[zi].z = 20;
  zr[zi].x = 0.0;
  zr[zi].hff = 1.25;
  zr[zi].haf = 1.0;
  zr[zi].al = 20.0;
  omega[zi] = 1;
  farbe[zi] = 1;

  float abst = 0.0,
        w = 0.0;

  int nz = 0;

  d_handle = fopen(d_name, "r");
  if(d_handle == NULL) {
    printf("Fehler beim ffnen der Datei\n");
    exit(1);
  }

  char zeile[120],      // Zeile
       *ende;

  ende = fgets(zeile, 120, d_handle);
  while(ende != NULL) {
    // Auswerten
    char *s = zeile;

    // Kommentar abschneiden, falls vorhanden
    while(*s != '%' && *s != '\0')
      s++;
    *s = '\0';

    s = zeile;
    // trennendes "=" suchen; gleichzeitig Kleinbuchstaben bilden
    while(*s != '=' && *s != '\0') {
      if(*s >= 'A' && *s <= 'Z')
        (*s) |= 0x20;
      s++;
    }
    if(*s == '=') {   // ein = liegt vor => interpretieren
      *s = '\0';      // trennen
      s++;
      if(!strcmp(zeile, "lagex"))
        sscanf(s, "%f", &zr[zi].lx);
      else if(!strcmp(zeile, "lagey"))
        sscanf(s, "%f", &zr[zi].ly);
      else if(!strcmp(zeile, "lage") || !strcmp(zeile, "position"))
        sscanf(s, "%f%f", &zr[zi].lx, &zr[zi].ly);
      else if(!strcmp(zeile, "stellung"))
        sscanf(s, "%f", &zr[zi].lal);
      else if(!strcmp(zeile, "modul") || !strcmp(zeile, "m"))
        sscanf(s, "%f", &zr[zi].m);
      else if(!strcmp(zeile, "z"))
        sscanf(s, "%d", &zr[zi].z);
      else if(!strcmp(zeile, "x") || !strcmp(zeile, "korr"))
        sscanf(s, "%f", &zr[zi].x);
      else if(!strcmp(zeile, "hf")) {
        sscanf(s, "%f", &zr[zi].hff);
        zr[zi].hff /= zr[zi].m;
      } else if(!strcmp(zeile, "ha")) {
        sscanf(s, "%f", &zr[zi].haf);
        zr[zi].haf /= zr[zi].m;
      } else if(!strcmp(zeile, "hff"))
        sscanf(s, "%f", &zr[zi].hff);
      else if(!strcmp(zeile, "haf"))
        sscanf(s, "%f", &zr[zi].haf);
      else if(!strcmp(zeile, "eingriff") || !strcmp(zeile, "alpha") ||
      !strcmp(zeile, ""))
        sscanf(s, "%f", &zr[zi].al);
      else if(!strcmp(zeile, "drehzahl") || !strcmp(zeile, "n") ||
      !strcmp(zeile, "omega") || !strcmp(zeile, "speed"))
        sscanf(s, "%d", &omega[zi]);
      else if(!strcmp(zeile, "a") || !strcmp(zeile, "abstand"))
        sscanf(s, "%f", &abst);
      else if(!strcmp(zeile, "nz"))
        sscanf(s, "%d", &nz);
      else if(!strcmp(zeile, "winkel") || !strcmp(zeile, "w"))
        sscanf(s, "%f", &w);
      else if(!strcmp(zeile, "farbe"))
        sscanf(s, "%d", &farbe[zi]);
      else if(!strcmp(zeile, "anrad")) {
        int ba;
        sscanf(s, "%d", &ba);
        ba--;  // weil von 0 beginnend
        zi++;
        zr[zi].berechne_neulage(zr[ba], w, nz, abst);
        farbe[zi] = farbe[ba];
        omega[zi] = -omega[ba];
        abst = 0.0;
        nz = 0;
      } else {
        printf("Nicht verstndliche Option:\n%s\n", zeile);
        fclose(d_handle);
        exit(1);
      }
    }
    ende = fgets(zeile, 120, d_handle);  // Nchste Zeile einlesen
  }
  fclose(d_handle);
  zi++;

  // Limiten berechnen
  e_fl = 1;
  for(register int i=0; i < zi; i++)
    zr[i].aktualisiere_limiten();

  // Limiten auf 4:3 anpassen
  if(y_x_verhaeltnis * (x_max - x_min) > (y_max - y_min)) {
    register float h = ldexp(y_min + y_max, -1),
                   h2 = ldexp(x_max - x_min, -1) * y_x_verhaeltnis;
    y_min1 = h - h2;
    y_max1 = h + h2;
    x_min1 = x_min;
    x_max1 = x_max;
  } else {
    register float h = ldexp(x_min + x_max, -1),
                   h2 = ldexp(y_max - y_min, -1) / y_x_verhaeltnis;
    x_min1 = h - h2;
    x_max1 = h + h2;
    y_min1 = y_min;
    y_max1 = y_max;
  }

  // Einstellen des Grafikmodus
  printf("Wichtiger Hinweis: ltere Grafiktreiber knnen Sie nur ber die\n"
  "Treiberkonstante ansteuern, ndern Sie diese daher nur ab, falls auf\n"
  "Ihrem System die Direkt-Videomoduswahl nicht korrekt luft.\n\n");
  lies_interv_ganze_zahl("Grafiktreiberkonstante Ihres Systems", "/",
  GR_320_200_graphics, GR_width_height_color_graphics, &treiber_konst);

  lies_positive_zahl("Horizontale Auflsung (Anzahl Bildpunkte)", "/",
  &x_aufl);
  y_aufl = int(float(x_aufl) * y_x_verhaeltnis + .9); // Addition von 0.9
  // nur, um Vorgabewert 0 zu vermeiden
  lies_positive_zahl("Vertikale Auflsung (Anzahl Bildpunkte)", "/",
  &y_aufl);
  lies_positive_zahl("Anzahl Farben (log2(Eingabe) ganzzahlig)", "/",
  &a_far);

  printf("\nEingabe der Parameter fr die Grafik\n");
  // Anzahl Bilder einlesen
  do {
    a_bilder = 1;
    lies_positive_zahl("Anzahl Bilder fr Animation (1=Standbild)", "/",
    &a_bilder);
    if(a_bilder > max_bilder)
      printf("Eingegebene Anzahl berschreitet Speicherkapazitt!\n");
  } while(a_bilder > max_bilder);

  // Alter Videomodus sichern
  alter_videomodus = GrCurrentMode(); // alter Videomodus sichern

  GrSetMode(treiber_konst, x_aufl, y_aufl, a_far);
  // Effektive Werte auslesen
  x_aufl = GrScreenX();
  y_aufl = GrScreenY();
  a_far = GrNumColors();

  GrClearScreen(0);
  setze_bereich(x_min1, x_max1, y_min1, y_max1);

  // Bilder berechnen
  if(a_bilder == 1)     // Nur Standbild
    for(int i=0; i < zi; i++)
      zr[i].zeichne(farbe[i]);
  else {
    // Grsse eines Grafik-Contexts bestimmen
    int po_x = int(floor((x_min - x_ba) * fx)), // Grafikkontext-Plazierposi-
        po_y = int(floor((y_ba - y_max) * fy)), // tion, x und y
        breit_x = int(floor((x_max - x_ba) * fx)) - po_x, // Breite und Hhe
        breit_y = int(floor((y_ba - y_min) * fy)) - po_y; // des Contexts
    po_x++;
    po_y++;
    // Basis verschieben => ergibt gltiges Lokalkoordinatensystem
    // innerhalb der Berechnung im Grafik-Kontext
    x_ba += float(po_x) / fx;
    y_ba -= float(po_y) / fy;

    // Wartebalken fr Berechnung anzeigen
    int w_balk_h = y_aufl / 10 - 1, // Wartebalkenhhe
        w_balk_p = 0,               // Wartebalkenposition
        w_balk_zyk = 0,             // Wartebalkenzyklus, aktueller Stand
        w_balk_zyk_l = (zi * a_bilder) / x_aufl; // Wartebalkenzyklus, Lnge
    if(w_balk_zyk_l == 0)
      w_balk_zyk_l = 1; // Sollte im Minimum 1 betragen

    // Wartebalken zeichnen
    GrFilledBox(0, 0, (zi * a_bilder) / w_balk_zyk_l - 1, w_balk_h,
    a_far > 256 ? 0x3E0 : 10);
    int w_balk_far = a_far > 256 ? 0x7C00 : 4;

    // Berechnung der Einzelbilder
    GrContext *Bild[max_bilder]; // Speicherplatz-Handles fr Bilder
    for(int b_nr = 0; b_nr < a_bilder; b_nr++) {
      Bild[b_nr] = GrCreateContext(breit_x, breit_y, NULL, NULL);
      // Bildspeicher anlegen
      GrSetContext(Bild[b_nr]); // Grafikbefehl auf Bildspeicher umleiten
      GrClearContext(0);        // lschen
      for(int i=0; i < zi; i++) {
        zr[i].zeichne(farbe[i]);
        // Zahnstellung vorrcken
        zr[i].lal += float(omega[i] * 360) / float(a_bilder * zr[i].z);
        w_balk_zyk++;
        if(w_balk_zyk == w_balk_zyk_l) {
          GrSetContext(NULL);   // Balken auf den Bildschirm ausgeben
          GrVLine(w_balk_p, 0, w_balk_h, w_balk_far);
          w_balk_p++;
          w_balk_zyk = 0;
          GrSetContext(Bild[b_nr]); // und wieder in den Context zurck
        }
      }
    }

    // Anzeigen der Grafik
    b_nr = 0;
    GrSetContext(NULL); // wieder normaler Bildschirm
    GrClearScreen(0);

    while(!kbhit()) {
      GrBitBlt(NULL, po_x, po_y, Bild[b_nr], 0, 0, breit_x - 1, breit_y - 1,
      GrWRITE);
      b_nr++;
      if(b_nr == a_bilder)
        b_nr = 0;
    }

    // Speicherplatz wieder freigeben
    for(int i = 0; i < a_bilder; i++)
      GrDestroyContext(Bild[i]);
  }

  // Tastendruck "vernichten" / darauf warten
  if(getch() == '\0')
    getch();

  // Wieder in den ursprnglichen Modus zurck
  GrSetMode(alter_videomodus);
  printf("\e[2J\e[2;1HDie Darstellung erfolgte in einer effektiven Bild"
  "schirmauflsung von\n");
  printf("%dx%d Bildpunkten bei %d Farben\n\n", x_aufl, y_aufl, a_far);
}
